#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#undef ATOMINFO_C__
#include "atominfo.h"


static char *progn = "atominfo";


static int str_icmp(const char *s, const char *t)
{
  char  cs, ct;


  while (*s || *t)
  {
    cs = toupper(*s++);
    ct = toupper(*t++);
    if (cs < ct) return -1;
    if (cs > ct) return  1;
  }

  return 0;
}


static void ShowSF_IT92_CAA(const T_SF_IT92_CAA *CAA)
{
  int  i, n;

  n = sizeof CAA->a / sizeof (*CAA->a);

  fputs("  IT92-CAA", stdout);

  for (i = 0; i < n; i++)
    fprintf(stdout, " a%d b%d", i + 1, i + 1);

  fputs(" c\n   ", stdout);

  for (i = 0; i < (sizeof CAA->a / sizeof (*CAA->a)); i++)
    fprintf(stdout, " %.6g %.6g", CAA->a[i], CAA->b[i]);

  fprintf(stdout, " %.6g\n", CAA->c);
}


static void ShowSF_WK95_CAA(const T_SF_WK95_CAA *CAA)
{
  int  i, n;

  n = sizeof CAA->a / sizeof (*CAA->a);

  fputs("  WK95-CAA", stdout);

  for (i = 0; i < n; i++)
    fprintf(stdout, " a%d b%d", i + 1, i + 1);

  fputs(" c\n   ", stdout);

  for (i = 0; i < (sizeof CAA->a / sizeof (*CAA->a)); i++)
    fprintf(stdout, " %.6g %.6g", CAA->a[i], CAA->b[i]);

  fprintf(stdout, " %.6g\n", CAA->c);
}

static void usage(void)
{
  fprintf(stderr,
    "usage: %s [-IT92|WK95] [-ExactLabel] [-Anomalous] [-AnomalousHenke] AtomLabel\n"
    "                [Start_sinTovL [End_sinTovL [StepSize]]]\n",
    progn);
  exit(1);
}


int main(int argc, char *argv[])
{
  int     iarg, i, n;
  int     iSES;
  int     F_IT92, F_WK95, F_ExactLabel;
  char    *AtomLabel, xtrac;
  double  Start_stol, End_stol, StepSize;
  double  stol, sfIT92, sfWK95, val;

  const T_PSE          *ePSE;
  const T_SF_IT92_CAA  *SF_IT92;
  const T_SF_WK95_CAA  *SF_WK95;
  const T_AtomRadius   *AR;

  const T_NeutronBondSL_NN92 *BCSL;
  const T_Ef1f2_Henke *Ef1f2Henke ;
  const T_Ef1f2_Sasaki *Ef1f2Sasaki ;
  float energy;
  int F_AnomalousHenke,F_AnomalousSasaki;

  F_IT92 = 0;
  F_WK95 = 0;
  F_ExactLabel = 0;
  F_AnomalousHenke=0;
  F_AnomalousSasaki=0;
  AtomLabel = NULL;

  Start_stol = 0.;
  End_stol = 1.;
  StepSize = .05;

  iSES = -1;

  for (iarg = 1; iarg < argc; iarg++)
  {
    if (argv[iarg][0] == '-' && isdigit(argv[iarg][1]) == 0)
    {
      if (str_icmp(argv[iarg], "-ExactLabel") == 0)
        F_ExactLabel = 1;
      else if (str_icmp(argv[iarg], "-IT92") == 0)
        F_IT92 = 1;
      else if (str_icmp(argv[iarg], "-WK95") == 0)
        F_WK95 = 1;
      else if (str_icmp(argv[iarg], "-Anomalous") == 0)  /*VFN*/
            F_AnomalousSasaki = 1;
      else if (str_icmp(argv[iarg], "-Sasaki") == 0)  /*VFN*/
            F_AnomalousSasaki = 1;
      else if (str_icmp(argv[iarg], "-Henke") == 0)  /*VFN*/
            F_AnomalousHenke = 1;

      else
        usage();
    }
    else if (AtomLabel == NULL)
    {
      AtomLabel = argv[iarg];
      iSES = 0;
    }
    else if (iSES >= 0 && iSES < 3)
    {
          n = sscanf(argv[iarg], "%lf %c", &val, &xtrac);
      if (n != 1 && ! (n == 2 && isspace(xtrac)))
        usage();

      if      (iSES == 0)
      {
            Start_stol = val;
        if (Start_stol < 0.) {
          fprintf(stderr, "%s: illegal parameter: Start_sinTovL\n", progn);
          usage();
        }
      }
      else if (iSES == 1)
      {
            End_stol = val;
        if (End_stol < 0. || End_stol >= 100.) {
          fprintf(stderr, "%s: illegal parameter: End_sinTovL\n", progn);
          usage();
        }
      }
      else
      {
            StepSize = val;
        if (StepSize <= 0.) {
          fprintf(stderr, "%s: illegal parameter: StepSize\n", progn);
          usage();
        }
      }

      iSES++;
    }
    else
      usage();
  }

  if (AtomLabel == NULL)
    usage();

  putc('#', stdout);
  for (i = 0; i < argc;)
    fprintf(stdout, " %s", argv[i++]);
  putc('\n', stdout);

  ePSE    = FindInPSE(AtomLabel, F_ExactLabel);
  SF_IT92 = FindSF_IT92_CAA(AtomLabel, F_ExactLabel);
  SF_WK95 = FindSF_WK95_CAA(AtomLabel, F_ExactLabel);
  AR      = FindAtomRadius(AtomLabel, F_ExactLabel);


  if (ePSE)
    fprintf(stdout, "%d \"%s\" \"%s\" %.3f\n",
      ePSE->Z, ePSE->Symbol, ePSE->Name, ePSE->Mass);
  else
    fprintf(stdout, "not found in PSE\n");

  fprintf(stdout,
    "Scattering Factor Label (IT92):");

  if (SF_IT92) {
    fprintf(stdout, " \"%s\"\n", SF_IT92->Label);
    ShowSF_IT92_CAA(SF_IT92);
  }
  else
    fprintf(stdout, " not found\n");

  fprintf(stdout,
    "Scattering Factor Label (WK95):");

  if (SF_WK95) {
    fprintf(stdout, " \"%s\"\n", SF_WK95->Label);
    ShowSF_WK95_CAA(SF_WK95);
  }
  else
    fprintf(stdout, " not found\n");

  fprintf(stdout, "Atom Radius");

  if (AR)
    fprintf(stdout, " %.2f \"%s\"\n", AR->Radius, AR->Label);
  else
    fprintf(stdout, " not found\n");

  if (F_IT92 == 0 && F_WK95 == 0) {
    F_IT92 = 1;
    F_WK95 = 1;
  }

  if ((SF_IT92 || SF_WK95) && iSES > 0)
  {
      fprintf(stdout, "   stol");

    if (SF_IT92 && F_IT92)
      fprintf(stdout, "     IT92");

    if (SF_WK95 && F_WK95)
      fprintf(stdout, "     WK95");

    putc('\n', stdout);

    sfIT92 = 0.;
    sfWK95 = 0.;

    for (i = 0;                                         /* ARBITRARY */
         (stol = Start_stol + i * StepSize) <= (End_stol + 1.e-4);
         i++
        )
    {
      fprintf(stdout, "  %5.3f", stol);

      if (SF_IT92 && F_IT92) {
        sfIT92 = CalcSF_IT92_CAA(SF_IT92, stol * stol);
        fprintf(stdout, "  %7.3f", sfIT92);
      }

      if (SF_WK95 && F_WK95) {
        sfWK95 = CalcSF_WK95_CAA(SF_WK95, stol * stol);
        fprintf(stdout, "  %7.3f", sfWK95);
      }

      putc('\n', stdout);
    }
  }

/*VFN Begin */
   BCSL    = FindNeutronBondSL_NN92(AtomLabel, F_ExactLabel);
   fprintf(stdout, "Neutron Bond scattering length:");
   if(BCSL)
      {
         fprintf(stdout, "   %.3f   \"%s\"\n", BCSL->BondCohScattLength, BCSL->Symbol);
         fprintf(stdout, "Neutron (2200m/s) cross section:   %.6f\n",
               BCSL->AbsCrossSect);
      }
      else fprintf(stdout, " not found\n");

   if(F_AnomalousSasaki)
   {
      Ef1f2Sasaki=FindEf1f2_Sasaki(AtomLabel, F_ExactLabel);
      fprintf(stdout, "XRay Anomalous Scattering factors (Sasaki): ");
      if(Ef1f2Sasaki)
      {
         fprintf(stdout,"\"%s\"\n",Ef1f2Sasaki->Symbol);
         fprintf(stdout," Energy(eV)      fprime        fsecond \n");

         for(i=0;; i++)
         {
            energy=Ef1f2Sasaki->Ef1f2List[i].Energy;
            if (energy == 0.) break;
            {
               fprintf(stdout, "    %.2f     %.4f      %.4f\n",
                  energy,
                  Ef1f2Sasaki->Ef1f2List[i].f1-ePSE->Z,
                  Ef1f2Sasaki->Ef1f2List[i].f2);
            }
         }
      }
      else fprintf(stdout, "not found\n");
   }
   if(F_AnomalousHenke)
   {
      Ef1f2Henke=FindEf1f2_Henke(AtomLabel, F_ExactLabel);
      fprintf(stdout, "XRay Anomalous Scattering factors (Henke): ");
      if(Ef1f2Henke)
      {
         fprintf(stdout,"\"%s\"\n",Ef1f2Henke->Symbol);
         fprintf(stdout," Energy(eV)      fprime        fsecond \n");

         for(i=0;; i++)
         {
            energy=Ef1f2Henke->Ef1f2List[i].Energy;
            if (energy == 0.) break;
            {
               fprintf(stdout, "    %.2f     %.4f      %.4f\n",
                  energy,
                  Ef1f2Henke->Ef1f2List[i].f1-ePSE->Z,
                  Ef1f2Henke->Ef1f2List[i].f2);
            }
         }
      }
      else fprintf(stdout, "not found\n");
   }
/*VFN End*/


  return 0;
}
